%{
/* GTKML - GTK Markup Language
 * Copyright (C) 1998 Scott Dybiec
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <gtk/gtk.h>

/* GTK and lex/yacc both define TRUE and FALSE */
#define GTKTRUE 1 
#define GTKFALSE 0
#undef TRUE
#undef FALSE

typedef struct Instance
{
   char *name;
   GtkWidget *value;
   struct Instance *next;
} Instance;

typedef struct Callback
{
   char *event;
   void (*action)();
   GtkWidget *widget;
} Callback;
    
typedef union
{
   int homogeneous;
   int spacing;
   int borderwidth;
   void (*justifyfunc)(); 
} Boxoption;

typedef struct BoxOptions 
{
   int homogeneous;
   int spacing;
   int borderwidth;
   void (*justify)(); 
} BoxOptions;
        
typedef struct ButtonOptions 
{
   char *tooltip;
   int borderwidth;
} ButtonOptions;
        
typedef struct ListItemOptions 
{
   char *tooltip;
   int borderwidth;
} ListItemOptions;
        
typedef struct FrameOptions 
{
   char *title;
   int frameshadow;
} FrameOptions;
        
typedef struct WindowOptions 
{
   char *title;
   int borderwidth;
} WindowOptions;
        
%}

%union {
    char             *string;       /* string buffer */
    int              sym;           /* integer value */
    GtkWidget        *widget;       /* GTK Widget */
    gint             number;        /* integers */
    GtkPositionType  tabposition;   /* notebook orientation */
    GSList           *radiogroup;   /* radio button group */
    Callback         *callback;     /* setup info for callbacks */
    void             (*cbfunc)();   /* callback function */
}

%token <string> QSTRING COMMA ID NUMBER
%token <sym> OBRACE CBRACE
%token <sym> FORM TITLE
%token <sym> HORIZONTALBOX VERTICALBOX
%token <sym> HOMOGENEOUS SPACING
%token <sym> FRAME
%token <sym> BORDERWIDTH 
%token <sym> BUTTON TOGGLEBUTTON CHECKBUTTON
%token <sym> LABEL
%token <sym> WINDOW
%token <sym> TRUE FALSE
%token <sym> TOOLTIP
%token <sym> ENTRY
%token <sym> NOTEBOOK PAGE 
%token <sym> TABPOSITION TOP BOTTOM LEFT RIGHT
%token <sym> RADIOBUTTON RADIOGROUP
%token <sym> LIST LISTITEM
%token <sym> FRAMESHADOW SHADOW_IN SHADOW_OUT SHADOW_NONE
%token <sym> SHADOW_ETCHED_IN SHADOW_ETCHED_OUT
%token <sym> WIDGET
%token <sym> ONCLICK ONPRESS ONRELEASE ONENTER ONLEAVE ONTOGGLE
%token <sym> SENSITIZE INSENSITIZE SHOW HIDE
%token <sym> TOGGLESENSITIVITY TOGGLEVISIBILITY
%token <sym> SUBMITFORM RESETFORM CANCELFORM
%token <sym> JUSTIFY POSITION CENTER
%token <sym> XPADDING YPADDING TABLE TABLEDATUM
%token <sym> COMBOBOX COMBOITEM
%token <sym> FILL


%type <string> title tooltip id optid qstring
%type <string> event
%type <widget> widget widgetinstance
%type <widget> boxblock frameblock label button
%type <widget> buttonblock windowblock labelblock
%type <widget> notebookblock notebook container
%type <widget> radiobuttonblock radiobutton radiogroupblock
%type <widget> box frame window entry entryblock
%type <widget> listblock list
%type <widget> listitemblock listitem
%type <widget> comboboxblock combobox
%type <widget> table tableblock
%type <number> spacing borderwidth maxlength 
%type <number> xpad ypad
%type <sym> homogeneous truefalse
%type <sym> frameshadow shadowtype
%type <cbfunc> justify position
%type <cbfunc> action
%type <tabposition> tabposition
%type <radiogroup> radiobuttons

%start widgetseries

%%

windowblock: window optid OBRACE windowoptionsblock widgets CBRACE
                     { gtk_widget_show ($1);
                       gtk_container_border_width (GTK_CONTAINER ($1), 
                                                   windowOptions.borderwidth);
                       gtk_window_set_title (GTK_WINDOW ($1), 
                                             windowOptions.title);
                       stack_pop (&parentstack);
                       $$ = $1;
                     }

window: WINDOW       { $$ = gtk_window_new (GTK_WINDOW_TOPLEVEL);
                       gtk_signal_connect (GTK_OBJECT ($$), "destroy",
                                           GTK_SIGNAL_FUNC (gtk_exit), NULL);
                       stack_push (&parentstack, (void *) $$);
                       windowOptions.title = NULL;
                       windowOptions.borderwidth = 0;
                     }

windowoptionsblock: /* empty */
	| windowoptions

windowoptions: windowoption
	| windowoptions windowoption

windowoption: title         { windowOptions.title = $1; }
 	| borderwidth       { windowOptions.borderwidth = $1; } 

optid:  /* empty */ { $$ = NULL; }
	| ID        { GtkWidget *parent;
                      parent = (GtkWidget *) stack_read_first (parentstack);
                      install_instance ($1, parent);
                      $$ = $1;
                    }

id:     ID         { $$ = $1; }

qstring: QSTRING   { $$ = $1; }

title: TITLE qstring	{ $$ = $2; }

widgets: /* empty */
	| widgetseries

widgetseries: widget    
             { gint parent_type;
               gint expand = GTKTRUE;
               gint fill = GTKFALSE;
               gint padding = 0;
               GtkWidget *parent;

               /* Is non-widget (radio group) or top-level define */
               if (($1 != NULL) && (parentstack != NULL)) 
               {
                  parent = (GtkWidget *) stack_read_first (parentstack);
                  parent_type = GTK_WIDGET_TYPE (parent);
                  if (GTK_IS_BOX(parent))
                     gtk_box_pack_start(GTK_BOX(parent), $1, 
                                        expand, fill, padding);
                  else
                     gtk_container_add (GTK_CONTAINER (parent), $1);
                  gtk_widget_show ($1);
               }
             }

	| widgetseries widget   
             { GtkWidget *parent;
               gint expand = GTKTRUE;
               gint fill = GTKFALSE;
               gint padding = 0;
               gint parent_type;

               /* Is non-widget (radio group) or top-level define */
               if (($2 != NULL) && (parentstack != NULL)) 
               {
                  parent = (GtkWidget *) stack_read_first (parentstack);
                  parent_type = GTK_WIDGET_TYPE (parent);
                  if (GTK_IS_BOX(parent))
                     gtk_box_pack_start(GTK_BOX(parent), $2, 
                                        expand, fill, padding);
                  else
                     gtk_container_add (GTK_CONTAINER (parent), $2);
                  gtk_widget_show ($2);
               }
             }

widget: boxblock
	| frameblock
	| labelblock
	| entryblock
	| buttonblock
	| notebookblock
	| radiogroupblock
	| listblock
	| tableblock
	| comboboxblock
	| windowblock
	| widgetinstance

widgetinstance: WIDGET id  { GtkWidget *widget;
                             Instance *inst;

                             inst = lookup_instance ($2);
                             if (inst != NULL)
                                $$ = inst->value;
                             else 
                             {
                                warning ("Undefined widget", $2);
                                $$ = NULL;
                             }
                             free ($2);
                           }

boxblock:  box optid OBRACE boxoptionsblock widgets CBRACE
            { guint i,j; 
              $$ = $1;
              gtk_box_set_homogeneous (GTK_BOX ($1), boxOptions.homogeneous);
              gtk_box_set_spacing (GTK_BOX ($1), boxOptions.spacing);
              if (boxOptions.justify == set_pack_end) { /* Reverse order */
                 GList *children;
                 GtkWidget *child;
                 guint num_children;
                 children = gtk_container_children (GTK_CONTAINER ($1)); 
                 num_children = g_list_length (children);
                 if (num_children > 0) {
                    for (i = 0; i < num_children - 1; i++) {
                       child = (g_list_nth (children, num_children-i-1))->data;
                       gtk_box_reorder_child (GTK_BOX ($1), 
                                              GTK_WIDGET (child),
                                              i);
                    }
                 }
                 g_list_free (children);
              }
              gtk_container_foreach(GTK_CONTAINER($1), (void *)(unset_fill), $1); 
              gtk_container_foreach(GTK_CONTAINER($1), boxOptions.justify, $1); 
              gtk_container_border_width (GTK_CONTAINER ($1), 
                                          boxOptions.borderwidth);
              stack_pop (&parentstack);
            }

boxoptionsblock: /* empty */
	| boxoptions

boxoptions: boxoption
	| boxoptions boxoption

boxoption: homogeneous { boxOptions.homogeneous = $1; }
	| spacing      { boxOptions.spacing = $1; }
	| borderwidth  { boxOptions.borderwidth = $1; }
	| justify      { boxOptions.justify = $1; }

justify: JUSTIFY position  { $$ = $2; }

truefalse: TRUE     { $$ = GTKTRUE; }
	| FALSE     { $$ = GTKFALSE; }

homogeneous: HOMOGENEOUS truefalse  { $$ = $2; }

spacing: SPACING NUMBER  { $$ = atoi($2); }

box: HORIZONTALBOX          { $$ = gtk_hbox_new (GTKFALSE, 0);
                              stack_push (&parentstack, (void *) $$);
                              boxOptions.homogeneous = GTKFALSE;
                              boxOptions.spacing = 0;
                              boxOptions.borderwidth = 0;
                              boxOptions.justify = set_pack_center;
			    }
	| VERTICALBOX       { $$ = gtk_vbox_new (GTKFALSE, 0);
                              stack_push (&parentstack, (void *) $$);
                              boxOptions.homogeneous = GTKFALSE;
                              boxOptions.spacing = 0;
                              boxOptions.borderwidth = 0;
                              boxOptions.justify = set_pack_center;
		            }

position: TOP     { 
                    GtkWidget *parent;
                    parent = (GtkWidget *) stack_read_first (parentstack);
                    if (GTK_IS_HBOX (parent))
                       warning("Use Left or Right for HorizontalBox", NULL);
                    $$ = set_pack_start; 
                  }
        | BOTTOM  { 
                    GtkWidget *parent;
                    parent = (GtkWidget *) stack_read_first (parentstack);
                    if (GTK_IS_HBOX (parent))
                       warning("Use Left or Right for HorizontalBox", NULL);
                    $$ = set_pack_end; 
                  }
        | LEFT    {
                    GtkWidget *parent;
                    parent = (GtkWidget *) stack_read_first (parentstack);
                    if (GTK_IS_VBOX (parent))
                       warning("Use Top or Bottom for VerticalBox", NULL);
                    $$ = set_pack_start; 
                  }
        | RIGHT   { 
                    GtkWidget *parent;
                    parent = (GtkWidget *) stack_read_first (parentstack);
                    if (GTK_IS_VBOX (parent))
                       warning("Use Top or Bottom for VerticalBox", NULL);
                    $$ = set_pack_end; 
                  }
        | CENTER  { $$ = set_pack_center; }
        | FILL    { $$ = set_fill; }

frameblock: frame optid OBRACE frameoptionsblock widgets CBRACE
                   { gtk_frame_set_label (GTK_FRAME ($1), 
                                          frameOptions.title);
                     gtk_frame_set_shadow_type (GTK_FRAME ($1), 
                                                frameOptions.frameshadow);
                     stack_pop (&parentstack);
                   }

frame: FRAME       { $$ = gtk_frame_new (NULL);
                     stack_push (&parentstack, (void *) $$);
                     frameOptions.title = NULL;
                     frameOptions.frameshadow = GTK_SHADOW_ETCHED_IN;
                   }

frameoptionsblock: /* empty */
	| frameoptions

frameoptions: frameoption
	| frameoptions frameoption

frameoption: title          { frameOptions.title = $1; }
 	| frameshadow       { frameOptions.frameshadow = $1; } 

frameshadow: FRAMESHADOW shadowtype   { $$ = $2; }

shadowtype: SHADOW_IN         { $$ = GTK_SHADOW_IN; }
	| SHADOW_OUT          { $$ = GTK_SHADOW_OUT; }
	| SHADOW_ETCHED_IN    { $$ = GTK_SHADOW_ETCHED_IN; }
	| SHADOW_ETCHED_OUT   { $$ = GTK_SHADOW_ETCHED_OUT; }
	| SHADOW_NONE         { $$ = GTK_SHADOW_NONE; }

borderwidth: BORDERWIDTH NUMBER  { $$ = atoi($2); }

labelblock: label optid qstring   { gtk_label_set (GTK_LABEL($1), $3);
                                    stack_pop (&parentstack);
                                    $$ = $1;
                                  }

label: LABEL   { $$ = gtk_label_new ("Default Label"); 
                 stack_push (&parentstack, (void *) $$);
               }

entryblock: entry optid maxlength   
              { 
                $$ = $1;
                if ($3 != -1) gtk_entry_set_max_length (GTK_ENTRY ($$), $3);
                stack_pop (&parentstack);
              }

entry: ENTRY   { $$ = gtk_entry_new ();
                 stack_push (&parentstack, (void *) $$);
               }

maxlength: /* empty */  { $$ = -1; }
	| NUMBER        { $$ = atoi($1); }

buttonblock: button optid qstring   
                 { GtkWidget *label_widget; 
                   label_widget = gtk_label_new ($3);
                   gtk_misc_set_alignment (GTK_MISC (label_widget), 0.5, 0.5);
                   gtk_container_add (GTK_CONTAINER ($1), label_widget);
                   gtk_widget_show (label_widget);
                   stack_pop(&parentstack);
                   $$ = $1; 
                 }

	| button optid OBRACE buttonoptionsblock widgets CBRACE  
                    { GtkTooltips *tooltip;

                      if (buttonOptions.tooltip != NULL) {
                         tooltip = gtk_tooltips_new();
                         gtk_tooltips_set_tip (tooltip, $1, 
                                               buttonOptions.tooltip, NULL);
                      }
                      gtk_container_border_width (GTK_CONTAINER ($1), 
                                                  buttonOptions.borderwidth);
                      stack_pop (&parentstack);
                      $$ = $1;
                    }

button: BUTTON              { $$ = gtk_button_new ();
                              stack_push (&parentstack, (void *) $$);
                              buttonOptions.borderwidth = 0;
                              buttonOptions.tooltip = NULL;
                            }
	| TOGGLEBUTTON      { $$ = gtk_toggle_button_new ();
                              stack_push (&parentstack, (void *) $$);
                              buttonOptions.borderwidth = 0;
                              buttonOptions.tooltip = NULL;
                            }
	| CHECKBUTTON       { $$ = gtk_check_button_new ();
                              stack_push (&parentstack, (void *) $$);
                              buttonOptions.borderwidth = 0;
                              buttonOptions.tooltip = NULL;
                            }

buttonoptionsblock: /* empty */
	| buttonoptions

buttonoptions: buttonoption
	| buttonoptions buttonoption

buttonoption: tooltip  { buttonOptions.tooltip = $1; }
 	| borderwidth  { buttonOptions.borderwidth = $1; } 
	| signal

signal: event action id  
                  { GtkWidget *source;
                    Instance *inst;
                    guint type;

                    source = stack_read_first (parentstack);
                    inst = lookup_instance ($3);

                    if (inst != NULL)
                    {
                       type = gtk_signal_lookup ($1, GTK_OBJECT_TYPE (source));
                       if (type)
                       {
                          gtk_signal_connect (GTK_OBJECT (source), $1,
                                                GTK_SIGNAL_FUNC ($2),
                                                inst->value);
                       }
                       else
                          warning("Widget cannot emit event", $1);
                    }
                    else
                    {
                       warning ("Undefined widget", $3);
                    }
                  }
                              

event: ONCLICK           { $$ = "clicked"; }
	| ONPRESS        { $$ = "pressed"; }
	| ONRELEASE      { $$ = "released"; }
	| ONENTER        { $$ = "enter"; }
	| ONLEAVE        { $$ = "leave"; }
	| ONTOGGLE       { $$ = "toggled"; }

action: SENSITIZE            { $$ = widget_set_sensitive; }
	| INSENSITIZE        { $$ = widget_set_insensitive; }
	| TOGGLESENSITIVITY  { $$ = widget_toggle_sensitivity; }
	| SHOW               { $$ = widget_show; }
	| HIDE               { $$ = widget_hide; }
	| TOGGLEVISIBILITY   { $$ = widget_toggle_visibility; }
	| SUBMITFORM         { $$ = widget_show; }
	| RESETFORM          { $$ = widget_show; }
	| CANCELFORM         { $$ = widget_show; }

 
tooltip: TOOLTIP qstring  { $$ = $2; }

notebookblock: notebook optid OBRACE tabposition pages CBRACE
                     { gtk_notebook_set_tab_pos (GTK_NOTEBOOK ($1), $4);
                       stack_pop (&parentstack);
                       gtk_widget_show ($1);
                       $$ = $1;
                     }

notebook: NOTEBOOK      { $$ = gtk_notebook_new();
                          stack_push (&parentstack, (void *) $$);
                        }
pages: pageblock           
	| pages pageblock 
	;

pageblock: PAGE OBRACE title container CBRACE
          { GtkWidget *label;
            GtkWidget *parent;

            gtk_widget_show ($4);
            label = gtk_label_new ($3);
            parent = (GtkWidget *) stack_read_first (parentstack);
            gtk_notebook_append_page (GTK_NOTEBOOK (parent), $4, label);
          }

container: boxblock      { $$ = $1; }
	| notebookblock  { $$ = $1; }
	| frameblock     { $$ = $1; }

tabposition: /* empty */      { $$ = GTK_POS_TOP; }
	| TABPOSITION TOP     { $$ = GTK_POS_TOP; }
	| TABPOSITION BOTTOM  { $$ = GTK_POS_BOTTOM; }
	| TABPOSITION LEFT    { $$ = GTK_POS_LEFT; }
	| TABPOSITION RIGHT   { $$ = GTK_POS_RIGHT; }

radiogroupblock: radiogroup OBRACE radiobuttons CBRACE
                    { 
                      stack_pop (&parentstack);
                      $$ = NULL;
                    }

radiogroup: RADIOGROUP   { stack_push (&parentstack, (void *) NULL); }

radiobuttons: 
	radiobuttonblock 
                 { GSList *group;

                   $$ = gtk_radio_button_group (GTK_RADIO_BUTTON ($1));
                   group = (GSList *) stack_read_first (parentstack);
                   stack_pop (&parentstack);
                   stack_push (&parentstack, (void *) $$);
                 }
	| radiobuttons radiobuttonblock
                 { GSList *group;

                   $$ = gtk_radio_button_group (GTK_RADIO_BUTTON ($2));
                   group = (GSList *) stack_read_first (parentstack);
                   stack_pop (&parentstack);
                   stack_push (&parentstack, (void *) $$);
                 }

radiobuttonblock: 
	radiobutton optid qstring  
             { GSList *group;
               GtkWidget *parent;

               stack_pop (&parentstack);
               group = (GSList *) stack_read_first (parentstack);
               parent = (GtkWidget *) stack_read_nth (parentstack, 1);
               $$ = gtk_radio_button_new_with_label (group, $3); 
               gtk_container_add (GTK_CONTAINER (parent), $$);
               gtk_widget_show ($$);
             }

	| radiobutton optid OBRACE radiobuttonoptionsblock widgets CBRACE  
             { GtkTooltips *tooltip;
               GtkWidget *parent;

               if (radioButtonOptions.tooltip != NULL) {
                  tooltip = gtk_tooltips_new();
                  gtk_tooltips_set_tip (tooltip, $1, 
                                        radioButtonOptions.tooltip, NULL);
               }
               gtk_container_border_width (GTK_CONTAINER ($1), 
                                           radioButtonOptions.borderwidth);
               
               stack_pop (&parentstack);
               parent = (GtkWidget *) stack_read_nth (parentstack, 1);
               gtk_container_add (GTK_CONTAINER (parent), $1);
               gtk_widget_show ($1);
               $$ = $1;
             }
                            
radiobutton: RADIOBUTTON
                    { GSList *group;

                      group = (GSList *) stack_read_first (parentstack);
                      $$ = gtk_radio_button_new (group);
                      stack_push (&parentstack, (void *) $$);
                      radioButtonOptions.borderwidth = 0;
                      radioButtonOptions.tooltip = NULL;
                    }

radiobuttonoptionsblock: /* empty */
	| radiobuttonoptions

radiobuttonoptions: radiobuttonoption
	| radiobuttonoptions radiobuttonoption

radiobuttonoption: tooltip  { radioButtonOptions.tooltip = $1; }
 	| borderwidth       { radioButtonOptions.borderwidth = $1; } 

listblock: list optid OBRACE listitems CBRACE
                    {  stack_pop (&parentstack);
                       stack_pop (&parentstack);
                      $$ = $1;
                    }

list: LIST     { GtkWidget *list;

                 $$ = gtk_scrolled_window_new (NULL, NULL);
                 list = gtk_list_new();
                 gtk_container_add (GTK_CONTAINER ($$), list);
                 gtk_widget_show (list);
                 stack_push (&parentstack, (void *) list);
                 stack_push (&parentstack, (void *) $$);
               }

listitems: listitemblock            
	| listitems listitemblock  

listitemblock: 
	listitem optid qstring  
             { GtkWidget *label;
               GtkWidget *parent;

               stack_pop (&parentstack);
               label = gtk_label_new ($3);
               gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
               gtk_container_add (GTK_CONTAINER ($1), label);
               gtk_widget_show (label);
               parent = (GtkWidget *) stack_read_nth (parentstack, 1); 
               gtk_container_add (GTK_CONTAINER (parent), $1);
               gtk_widget_show ($1);
               $$ = $1;
             }

	| listitem optid OBRACE listitemoptionsblock widgets CBRACE  
             { GtkWidget *parent;

               gtk_container_border_width (GTK_CONTAINER ($1), 
                                           buttonOptions.borderwidth);
               stack_pop (&parentstack);
               parent = (GtkWidget *) stack_read_nth (parentstack, 1);
               gtk_container_add (GTK_CONTAINER (parent), $1);
               gtk_widget_show ($1);
               $$ = $1;
             }
                            
listitem: LISTITEM
             { $$ = gtk_list_item_new ();
               stack_push (&parentstack, (gpointer) $$);
               listItemOptions.borderwidth = 0;
               listItemOptions.tooltip = NULL;
             }

listitemoptionsblock: /* empty */
	| listitemoptions

listitemoptions: listitemoption
	| listitemoptions listitemoption

listitemoption: tooltip     { listItemOptions.tooltip = $1; }
 	| borderwidth       { listItemOptions.borderwidth = $1; } 

tableblock: table optid OBRACE tableelements CBRACE
             {
               $$ = $1;
               gtk_widget_show ($1);
               stack_pop(&parentstack);
             }

table: TABLE NUMBER NUMBER
             { 
               $$ = gtk_table_new (atoi($2), atoi($3), GTKFALSE);
               stack_push (&parentstack, (gpointer) $$);
             }

tableelements: /* empty */
	| tabledata

tabledata: tabledatumblock
	| tabledata tabledatumblock

tabledatumblock: TABLEDATUM NUMBER NUMBER NUMBER NUMBER OBRACE xpad ypad widget CBRACE
             {
               GtkWidget *table;
               table = (GtkWidget *) stack_read_first (parentstack);
               gtk_table_attach (GTK_TABLE (table), $9, 
                                 atoi($2), atoi($3), atoi($4), atoi($5), 
                                 GTK_FILL, GTK_FILL, 
                                 $7, $8);
               gtk_widget_show ($9);
             }

xpad: /* empty */          { $$ = 0; }
	| XPADDING NUMBER  { $$ = atoi($2); }

ypad: /* empty */          { $$ = 0; }
	| YPADDING NUMBER  { $$ = atoi($2); }

comboboxblock: combobox optid OBRACE comboitems CBRACE
                    { GtkWidget *combo;
                      GList *items;
                      stack_pop (&parentstack);
                      items = stack_pop (&parentstack);
                      gtk_combo_set_popdown_strings (GTK_COMBO (combo), items);
                      $$ = combo;
                    }
                             

combobox: COMBOBOX  { 
                      GList *items = NULL;
                      stack_push (&parentstack, items);
                      $$ = gtk_combo_new(); 
                      stack_push (&parentstack, $$);
                    }

comboitems: comboitem
	| comboitems comboitem

comboitem: COMBOITEM qstring
                    { GList *items;
                      GtkWidget *combo;
                      combo = (GtkWidget *) stack_pop (&parentstack);
                      items = stack_pop (&parentstack);
                      items = g_list_append (items, $2);
                      stack_push (&parentstack, items);
                      stack_push (&parentstack, combo);
                    }


%%

char            *progname = "gtkml";
int             lineno = 1;
GList           *parentstack = NULL;
Instance        *instlist = NULL;
GList           *cbstack = NULL;
BoxOptions      boxOptions;
ButtonOptions   buttonOptions;
ButtonOptions   radioButtonOptions;
WindowOptions   windowOptions;
ListItemOptions listItemOptions;
FrameOptions    frameOptions;
char            *usage = "%s: usage [infile] [outfile]\n";

void set_pack_start (GtkWidget *child, GtkBox *box)
{
   gint expand;
   gint fill;
   gint padding;
   GtkPackType pack_type;

   gtk_box_query_child_packing (box, child, 
                                &expand, &fill, &padding, &pack_type);
   gtk_box_set_child_packing (box, child, 
                              GTKFALSE, fill, padding, GTK_PACK_START);
}

void set_pack_end (GtkWidget *child, GtkBox *box)
{
   gint expand;
   gint fill;
   gint padding;
   GtkPackType pack_type;

   gtk_box_query_child_packing (box, child, 
                                &expand, &fill, &padding, &pack_type);
   gtk_box_set_child_packing (box, child, 
                              GTKFALSE, fill, padding, GTK_PACK_END);
}

void set_pack_center (GtkWidget *child, GtkBox *box)
{
   gint expand;
   gint fill;
   gint padding;
   GtkPackType pack_type;

   gtk_box_query_child_packing (box, child, 
                                &expand, &fill, &padding, &pack_type);
   gtk_box_set_child_packing (box, child, 
                              GTKTRUE, fill, padding, GTK_PACK_START);
}

void set_fill (GtkWidget *child, GtkBox *box)
{
   gint expand;
   gint fill;
   gint padding;
   GtkPackType pack_type;

   gtk_box_query_child_packing (box, child, 
                                &expand, &fill, &padding, &pack_type);
   gtk_box_set_child_packing (box, child, 
                              GTKTRUE, GTKTRUE, padding, pack_type);
}

void unset_fill (GtkWidget *child, GtkBox *box)
{
   gint expand;
   gint fill;
   gint padding;
   GtkPackType pack_type;

   gtk_box_query_child_packing (box, child, 
                                &expand, &fill, &padding, &pack_type);
   gtk_box_set_child_packing (box, child, 
                              GTKFALSE, GTKFALSE, padding, pack_type);
}

void widget_show (GtkWidget *widget, gpointer *data)
{
   gtk_widget_show ((GtkWidget *) data);
}

void widget_hide (GtkWidget *widget, gpointer *data)
{
   gtk_widget_hide ((GtkWidget *) data);
}

void widget_set_sensitive (GtkWidget *widget, gpointer *data)
{
   gtk_widget_set_sensitive ((GtkWidget *) data, GTKTRUE);
}

void widget_set_insensitive (GtkWidget *widget, gpointer *data)
{
   gtk_widget_set_sensitive ((GtkWidget *) data, GTKFALSE);
}

void widget_toggle_sensitivity (GtkWidget *widget, gpointer *data)
{
   if (GTK_WIDGET_IS_SENSITIVE ((GtkWidget *) data) == GTKTRUE)
      gtk_widget_set_sensitive ((GtkWidget *) data, GTKFALSE);
   else
      gtk_widget_set_sensitive ((GtkWidget *) data, GTKTRUE);
}

void widget_toggle_visibility (GtkWidget *widget, gpointer *data)
{
   if (GTK_WIDGET_VISIBLE ((GtkWidget *) data) == GTKTRUE)
      gtk_widget_hide ((GtkWidget *) data);
   else
      gtk_widget_show ((GtkWidget *) data);
}


Instance *lookup_instance (char *inst)
{
   Instance *ip;

   for (ip = instlist; ip != NULL; ip = ip->next)
      if (strcmp (ip->name, inst) == 0)
         return ip;
   return NULL;        /* NULL ==> not found */
}

Instance *install_instance (char *name, GtkWidget *value)
{
   Instance *ip;

   ip = lookup_instance (name);
   if (ip != NULL)
      warning ("Redefining previously defined instance", ip->name);
   else
   {
      ip = (Instance *) malloc (sizeof (Instance));
      ip->name = name;
      ip->next = instlist;
      instlist = ip;
   }
   ip->value = value;
   return ip;
}


int stack_push (GList **stack, void *data)
{
	*stack = g_list_prepend (*stack, (gpointer) data);
}

void *stack_pop (GList **stack)
{
      void *head;

      head = (g_list_first (*stack))->data; 
      *stack = g_list_remove (*stack, head);
      return (head);
}

void *stack_read_nth (GList *stack, int nth)
{
	void *element;

	element = (g_list_nth (stack, nth))->data;
	return (element);
}
	
void *stack_read_first (GList *stack)
{
	void *element;

	element = (g_list_first (stack))->data;
	return (element);
}
	

/* 
yyerror(const char *msg)
{
	fprintf(stderr, "%d: %s at %s\n", lineno, msg, yytext);
}
*/

main(int argc, char *argv[])
{
	char *outfile;
	char *infile;
	extern FILE *gtkmlin, *gtkmlout;
	yydebug = 1;
    
	progname = argv[0];

	gtk_init(&argc, &argv);

	if(argc > 2)
	{
        	fprintf(stderr,usage, progname);
		exit(1);
	}
	if(argc > 1)
	{
		infile = argv[1];
		/* open for read */
		gtkmlin = fopen(infile,"r");
		if(gtkmlin == NULL) /* open failed */
		{
			fprintf(stderr,"%s: cannot open %s\n", 
				progname, infile);
			exit(1);
		}
	}

	/* yyout = fopen(outfile,"w"); */
    
	/* normal interaction on gtkmlin and 
	   yyout from now on */
    
	if (yyparse())
	   exit(0); 
        else
           gtk_main ();
    
}

warning(char *s, char *t) /* print warning message */
{
	fprintf(stderr, "%s warning: %s", progname, s);
	if (t)
		fprintf(stderr, " \"%s\"", t);
	fprintf(stderr, " line %d\n", lineno);
}
